import torch


class CollatorWithPadding:
    def __init__(
        self,
        max_features: int,
        pad_to_max_features: bool,
    ) -> None:
        self.max_features = max_features
        self.pad_to_max_features = pad_to_max_features

    def __call__(self, batch: list[dict[str, torch.Tensor]]) -> dict[str, torch.Tensor]:
        max_support_samples = max(dataset["x_support"].shape[0] for dataset in batch)
        max_query_samples = max(dataset["x_query"].shape[0] for dataset in batch)
        max_features = max(dataset["x_support"].shape[1] for dataset in batch)

        if self.pad_to_max_features:
            max_features = self.max_features

        batch_size = len(batch)

        tensor_dict = {
            "x_support": torch.zeros(
                (batch_size, max_support_samples, max_features), dtype=batch[0]["x_support"].dtype
            ),
            "y_support": torch.full(
                (batch_size, max_support_samples), fill_value=-100, dtype=batch[0]["y_support"].dtype
            ),
            "x_query": torch.zeros((batch_size, max_query_samples, max_features), dtype=batch[0]["x_query"].dtype),
            "y_query": torch.full((batch_size, max_query_samples), fill_value=-100, dtype=batch[0]["y_query"].dtype),
            "padding_features": torch.ones((batch_size, max_features), dtype=torch.bool),
            "padding_obs_support": torch.ones((batch_size, max_support_samples), dtype=torch.bool),
            "padding_obs_query": torch.ones((batch_size, max_query_samples), dtype=torch.bool),
        }

        for i, dataset in enumerate(batch):
            tensor_dict["x_support"][i, : dataset["x_support"].shape[0], : dataset["x_support"].shape[1]] = dataset[
                "x_support"
            ]
            tensor_dict["y_support"][i, : dataset["y_support"].shape[0]] = dataset["y_support"]
            tensor_dict["x_query"][i, : dataset["x_query"].shape[0], : dataset["x_support"].shape[1]] = dataset[
                "x_query"
            ]
            tensor_dict["y_query"][i, : dataset["y_query"].shape[0]] = dataset["y_query"]
            tensor_dict["padding_features"][i, : dataset["x_support"].shape[1]] = False
            tensor_dict["padding_obs_support"][i, : dataset["x_support"].shape[0]] = False
            tensor_dict["padding_obs_query"][i, : dataset["x_query"].shape[0]] = False

        return tensor_dict



from typing import Optional

import torch


class CollatorWithPadding:
    def __init__(self, pad_to_n_support_samples: Optional[int]) -> None:
        self.pad_to_n_support_samples = pad_to_n_support_samples

    def __call__(self, batch: list[dict[str, torch.Tensor]]) -> dict[str, torch.Tensor]:
        if self.pad_to_n_support_samples is not None:
            assert all(dataset["x_support"].shape[0] <= self.pad_to_n_support_samples for dataset in batch)
            self.n_support_samples = self.pad_to_n_support_samples
        else:
            self.n_support_samples = max(dataset["x_support"].shape[0] for dataset in batch)

        max_query_samples = max(dataset["x_query"].shape[0] for dataset in batch)

        n_support_features = batch[0]["x_support"].shape[1]
        n_query_features = batch[0]["x_query"].shape[1]
        y_dtype = batch[0]["y_support"].dtype

        batch_size = len(batch)

        tensor_dict = {
            "x_support": torch.zeros((batch_size, self.n_support_samples, n_support_features), dtype=torch.float32),
            "y_support": torch.zeros((batch_size, self.n_support_samples), dtype=y_dtype),
            "x_query": torch.zeros((batch_size, max_query_samples, n_query_features), dtype=torch.float32),
            "y_query": torch.zeros((batch_size, max_query_samples), dtype=y_dtype),
        }

        for i, dataset in enumerate(batch):
            tensor_dict["x_support"][i, : dataset["x_support"].shape[0], :] = dataset["x_support"]
            tensor_dict["y_support"][i, : dataset["y_support"].shape[0]] = dataset["y_support"]
            tensor_dict["x_query"][i, : dataset["x_query"].shape[0], :] = dataset["x_query"]
            tensor_dict["y_query"][i, : dataset["y_query"].shape[0]] = dataset["y_query"]

        return tensor_dict



# State-of-the-art for datasets < 100k samples. Requires a GPU with at least 20 GB VRAM.
hyperparameter_portfolio_zeroshot_gpu_2025_12_18 = {
    "TABDPT": [
        {
            "ag_args": {"name_suffix": "_c1", "priority": -3},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": False},
        },
        {
            "ag_args": {"name_suffix": "_r20", "priority": -5},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": False},
            "clip_sigma": 8,
            "feature_reduction": "subsample",
            "missing_indicators": False,
            "normalizer": "quantile-uniform",
            "permute_classes": False,
            "temperature": 0.5,
        },
        {
            "ag_args": {"name_suffix": "_r1", "priority": -7},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": False},
            "clip_sigma": 16,
            "feature_reduction": "subsample",
            "missing_indicators": False,
            "normalizer": "log1p",
            "permute_classes": False,
            "temperature": 0.5,
        },
        {
            "ag_args": {"name_suffix": "_r15", "priority": -9},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": False},
            "clip_sigma": 16,
            "feature_reduction": "subsample",
            "missing_indicators": False,
            "normalizer": "standard",
            "permute_classes": True,
            "temperature": 0.7,
        },
        {
            "ag_args": {"name_suffix": "_r22", "priority": -11},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": False},
            "clip_sigma": 8,
            "feature_reduction": "pca",
            "missing_indicators": True,
            "normalizer": "robust",
            "permute_classes": False,
            "temperature": 0.5,
        },
    ],
    "TABICL": [{"ag_args": {"name_suffix": "_c1", "priority": -4}, "ag_args_ensemble": {"refit_folds": True}}],
    "MITRA": [
        {
            "ag_args": {"name_suffix": "_c1", "priority": -12},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
        }
    ],
    "TABM": [
        {
            "ag_args": {"name_suffix": "_r99", "priority": -13},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 880,
            "d_embedding": 24,
            "dropout": 0.10792355695428629,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0013641856391615784,
            "n_blocks": 5,
            "num_emb_n_bins": 16,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {"name_suffix": "_r124", "priority": -17},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 208,
            "d_embedding": 16,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.00042152744054701374,
            "n_blocks": 2,
            "num_emb_n_bins": 109,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.00014007839435474664,
        },
        {
            "ag_args": {"name_suffix": "_r69", "priority": -21},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 848,
            "d_embedding": 28,
            "dropout": 0.40215621636031007,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0010413640454559532,
            "n_blocks": 3,
            "num_emb_n_bins": 18,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {"name_suffix": "_r184", "priority": -24},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 864,
            "d_embedding": 24,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0019256819924656217,
            "n_blocks": 3,
            "num_emb_n_bins": 3,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {"name_suffix": "_r34", "priority": -26},
            "amp": False,
            "arch_type": "tabm-mini",
            "batch_size": "auto",
            "d_block": 896,
            "d_embedding": 8,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.002459175026451607,
            "n_blocks": 4,
            "num_emb_n_bins": 104,
            "num_emb_type": "pwl",
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0006299584388562901,
        },
    ],
    "GBM_PREP": [
        {
            "ag.prep_params": [
                [
                    [["ArithmeticFeatureGenerator", {}]],
                    [
                        ["CategoricalInteractionFeatureGenerator", {"passthrough": True}],
                        ["OOFTargetEncodingFeatureGenerator", {}],
                    ],
                ]
            ],
            "ag.prep_params.passthrough_types": {"invalid_raw_types": ["category", "object"]},
            "ag_args": {"name_suffix": "_r13", "priority": -14},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "bagging_fraction": 0.9923026236907,
            "bagging_freq": 1,
            "cat_l2": 0.014290368488,
            "cat_smooth": 1.8662939903973,
            "extra_trees": True,
            "feature_fraction": 0.5533919718605,
            "lambda_l1": 0.914411672958,
            "lambda_l2": 1.90439560009,
            "learning_rate": 0.0193225778401,
            "max_cat_to_onehot": 18,
            "min_data_in_leaf": 28,
            "min_data_per_group": 54,
            "num_leaves": 64,
        },
        {
            "ag.prep_params": [
                [
                    [["ArithmeticFeatureGenerator", {}]],
                    [
                        ["CategoricalInteractionFeatureGenerator", {"passthrough": True}],
                        ["OOFTargetEncodingFeatureGenerator", {}],
                    ],
                ]
            ],
            "ag.prep_params.passthrough_types": {"invalid_raw_types": ["category", "object"]},
            "ag_args": {"name_suffix": "_r41", "priority": -16},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "bagging_fraction": 0.7215411996558,
            "bagging_freq": 1,
            "cat_l2": 1.887369154362,
            "cat_smooth": 0.0278693980873,
            "extra_trees": True,
            "feature_fraction": 0.4247583287144,
            "lambda_l1": 0.1129800247772,
            "lambda_l2": 0.2623265718536,
            "learning_rate": 0.0074201920651,
            "max_cat_to_onehot": 9,
            "min_data_in_leaf": 15,
            "min_data_per_group": 10,
            "num_leaves": 8,
        },
        {
            "ag.prep_params": [
                [
                    [["ArithmeticFeatureGenerator", {}]],
                    [
                        ["CategoricalInteractionFeatureGenerator", {"passthrough": True}],
                        ["OOFTargetEncodingFeatureGenerator", {}],
                    ],
                ]
            ],
            "ag.prep_params.passthrough_types": {"invalid_raw_types": ["category", "object"]},
            "ag_args": {"name_suffix": "_r31", "priority": -18},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "bagging_fraction": 0.9591526242875,
            "bagging_freq": 1,
            "cat_l2": 1.8962346412823,
            "cat_smooth": 0.0215219089995,
            "extra_trees": False,
            "feature_fraction": 0.5791844062459,
            "lambda_l1": 0.938461750637,
            "lambda_l2": 0.9899852075056,
            "learning_rate": 0.0397613094741,
            "max_cat_to_onehot": 27,
            "min_data_in_leaf": 1,
            "min_data_per_group": 39,
            "num_leaves": 16,
        },
        {
            "ag.prep_params": [],
            "ag_args": {"name_suffix": "_r21", "priority": -20},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "bagging_fraction": 0.7111549514262,
            "bagging_freq": 1,
            "cat_l2": 0.8679131150136,
            "cat_smooth": 48.7244965504817,
            "extra_trees": False,
            "feature_fraction": 0.425140839263,
            "lambda_l1": 0.5140528525242,
            "lambda_l2": 0.5134051978198,
            "learning_rate": 0.0134375321277,
            "max_cat_to_onehot": 16,
            "min_data_in_leaf": 2,
            "min_data_per_group": 32,
            "num_leaves": 20,
        },
        {
            "ag.prep_params": [
                [
                    [["ArithmeticFeatureGenerator", {}]],
                    [
                        ["CategoricalInteractionFeatureGenerator", {"passthrough": True}],
                        ["OOFTargetEncodingFeatureGenerator", {}],
                    ],
                ]
            ],
            "ag.prep_params.passthrough_types": {"invalid_raw_types": ["category", "object"]},
            "ag_args": {"name_suffix": "_r17", "priority": -23},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "bagging_fraction": 0.9277474245702,
            "bagging_freq": 1,
            "cat_l2": 0.0731876168104,
            "cat_smooth": 0.1369210915339,
            "extra_trees": False,
            "feature_fraction": 0.6680440910385,
            "lambda_l1": 0.0125057410295,
            "lambda_l2": 0.7157181359874,
            "learning_rate": 0.0351342879995,
            "max_cat_to_onehot": 20,
            "min_data_in_leaf": 1,
            "min_data_per_group": 2,
            "num_leaves": 64,
        },
    ],
    "CAT": [{"ag_args": {"name_suffix": "_c1", "priority": -15}}],
    "GBM": [
        {
            "ag_args": {"name_suffix": "_r73", "priority": -19},
            "bagging_fraction": 0.7295548973583,
            "bagging_freq": 1,
            "cat_l2": 1.8025485263237,
            "cat_smooth": 59.6178463268351,
            "extra_trees": False,
            "feature_fraction": 0.8242607305914,
            "lambda_l1": 0.7265522905459,
            "lambda_l2": 0.3492160682092,
            "learning_rate": 0.0068803786367,
            "max_cat_to_onehot": 16,
            "min_data_in_leaf": 1,
            "min_data_per_group": 10,
            "num_leaves": 24,
        },
        {
            "ag_args": {"name_suffix": "_r37", "priority": -22},
            "bagging_fraction": 0.8096374561947,
            "bagging_freq": 1,
            "cat_l2": 1.6385754694703,
            "cat_smooth": 16.1922506671724,
            "extra_trees": True,
            "feature_fraction": 0.885927003286,
            "lambda_l1": 0.0430386950502,
            "lambda_l2": 0.2507506811761,
            "learning_rate": 0.0079622660542,
            "max_cat_to_onehot": 23,
            "min_data_in_leaf": 7,
            "min_data_per_group": 49,
            "num_leaves": 6,
        },
        {
            "ag_args": {"name_suffix": "_r162", "priority": -25},
            "bagging_fraction": 0.7552878818396,
            "bagging_freq": 1,
            "cat_l2": 0.0081083103544,
            "cat_smooth": 75.7373446363438,
            "extra_trees": False,
            "feature_fraction": 0.6171258454584,
            "lambda_l1": 0.1071522383181,
            "lambda_l2": 1.7882554584069,
            "learning_rate": 0.0229328987255,
            "max_cat_to_onehot": 24,
            "min_data_in_leaf": 23,
            "min_data_per_group": 2,
            "num_leaves": 125,
        },
        {
            "ag_args": {"name_suffix": "_r57", "priority": -27},
            "bagging_fraction": 0.8515739264605,
            "bagging_freq": 1,
            "cat_l2": 0.2263901847144,
            "cat_smooth": 1.7397457971767,
            "extra_trees": True,
            "feature_fraction": 0.6284015946887,
            "lambda_l1": 0.6935431676756,
            "lambda_l2": 1.7605230133162,
            "learning_rate": 0.0294830579218,
            "max_cat_to_onehot": 52,
            "min_data_in_leaf": 8,
            "min_data_per_group": 3,
            "num_leaves": 43,
        },
        {
            "ag_args": {"name_suffix": "_r33", "priority": -28},
            "bagging_fraction": 0.9625293420216,
            "bagging_freq": 1,
            "cat_l2": 0.1236875455555,
            "cat_smooth": 68.8584757332856,
            "extra_trees": False,
            "feature_fraction": 0.6189215809382,
            "lambda_l1": 0.1641757352921,
            "lambda_l2": 0.6937755557881,
            "learning_rate": 0.0154031028561,
            "max_cat_to_onehot": 17,
            "min_data_in_leaf": 1,
            "min_data_per_group": 30,
            "num_leaves": 68,
        },
    ],
    "REALTABPFN-V2": [
        {
            "ag_args": {"name_suffix": "_r13", "priority": -1},
            "ag_args_ensemble": {"model_random_seed": 104, "vary_seed_across_folds": True},
            "balance_probabilities": False,
            "inference_config/OUTLIER_REMOVAL_STD": 6,
            "inference_config/POLYNOMIAL_FEATURES": "no",
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": [None, "safepower"],
            "preprocessing/append_original": False,
            "preprocessing/categoricals": "numeric",
            "preprocessing/global": None,
            "preprocessing/scaling": ["squashing_scaler_default", "quantile_uni_coarse"],
            "softmax_temperature": 1.0,
            "zip_model_path": ["tabpfn-v2-classifier-finetuned-zk73skhh.ckpt", "tabpfn-v2-regressor-v2_default.ckpt"],
        },
        {
            "ag_args": {"name_suffix": "_r106", "priority": -2},
            "ag_args_ensemble": {"model_random_seed": 848, "vary_seed_across_folds": True},
            "balance_probabilities": False,
            "inference_config/OUTLIER_REMOVAL_STD": 6,
            "inference_config/POLYNOMIAL_FEATURES": "no",
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": [None],
            "preprocessing/append_original": True,
            "preprocessing/categoricals": "numeric",
            "preprocessing/global": "svd_quarter_components",
            "preprocessing/scaling": ["quantile_uni_coarse"],
            "softmax_temperature": 0.8,
            "zip_model_path": ["tabpfn-v2-classifier-finetuned-zk73skhh.ckpt", "tabpfn-v2-regressor-v2_default.ckpt"],
        },
        {
            "ag_args": {"name_suffix": "_r11", "priority": -6},
            "ag_args_ensemble": {"model_random_seed": 88, "vary_seed_across_folds": True},
            "balance_probabilities": True,
            "inference_config/OUTLIER_REMOVAL_STD": 6,
            "inference_config/POLYNOMIAL_FEATURES": 25,
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": [None],
            "preprocessing/append_original": True,
            "preprocessing/categoricals": "onehot",
            "preprocessing/global": "svd_quarter_components",
            "preprocessing/scaling": ["safepower", "quantile_uni"],
            "softmax_temperature": 0.7,
            "zip_model_path": ["tabpfn-v2-classifier-finetuned-zk73skhh.ckpt", "tabpfn-v2-regressor-v2_default.ckpt"],
        },
        {
            "ag_args": {"name_suffix": "_c1", "priority": -8},
            "ag_args_ensemble": {"model_random_seed": 0, "vary_seed_across_folds": True},
            "zip_model_path": ["tabpfn-v2-classifier-finetuned-zk73skhh.ckpt", "tabpfn-v2-regressor-v2_default.ckpt"],
        },
        {
            "ag_args": {"name_suffix": "_r196", "priority": -10},
            "ag_args_ensemble": {"model_random_seed": 1568, "vary_seed_across_folds": True},
            "balance_probabilities": False,
            "inference_config/OUTLIER_REMOVAL_STD": 12,
            "inference_config/POLYNOMIAL_FEATURES": "no",
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": ["kdi_alpha_1.0"],
            "preprocessing/append_original": False,
            "preprocessing/categoricals": "numeric",
            "preprocessing/global": None,
            "preprocessing/scaling": ["squashing_scaler_default"],
            "softmax_temperature": 1.25,
            "zip_model_path": ["tabpfn-v2-classifier-finetuned-zk73skhh.ckpt", "tabpfn-v2-regressor-v2_default.ckpt"],
        },
    ],
}



from __future__ import annotations

import logging
import time

import einops
import numpy as np
import torch
from numpy.random import Generator
from sklearn.base import BaseEstimator

from autogluon.core.metrics import Scorer

from ..config.config_run import ConfigRun
from ..data.dataset_finetune import DatasetFinetune, DatasetFinetuneGenerator
from ..data.preprocessor import Preprocessor
from ..results.prediction_metrics import PredictionMetrics
from .callbacks import Checkpoint, EarlyStopping, TrackOutput
from .collator import CollatorWithPadding
from .enums import Task
from .get_loss import get_loss
from .get_optimizer import get_optimizer
from .get_scheduler import get_scheduler
from .y_transformer import create_y_transformer

logger = logging.getLogger(__name__)


class TrainerFinetune(BaseEstimator):
    def __init__(
        self,
        cfg: ConfigRun,
        model: torch.nn.Module,
        n_classes: int,
        stopping_metric: Scorer,
        use_best_epoch: bool = True,
        compute_train_metrics: bool = False,
    ) -> None:
        self.cfg = cfg
        self.model = model
        self.model.to(self.cfg.device)
        self.n_classes = n_classes

        self.loss = get_loss(self.cfg.task)
        self.optimizer = get_optimizer(self.cfg.hyperparams, self.model)
        self.scheduler = get_scheduler(self.cfg.hyperparams, self.optimizer)
        self.use_best_epoch = use_best_epoch
        self.compute_train_metrics = compute_train_metrics

        self.early_stopping = EarlyStopping(patience=self.cfg.hyperparams["early_stopping_patience"])
        self.preprocessor = Preprocessor(
            use_quantile_transformer=self.cfg.hyperparams["use_quantile_transformer"],
            use_feature_count_scaling=self.cfg.hyperparams["use_feature_count_scaling"],
            max_features=self.cfg.hyperparams["n_features"],
            task=self.cfg.task,
        )

        self.stopping_metric = stopping_metric
        self.best_epoch = None

    def set_random_seed(self) -> Generator:
        torch.manual_seed(self.cfg.seed)
        np.random.seed(self.cfg.seed)
        rng = np.random.default_rng(seed=self.cfg.seed)
        return rng

    def reset_optimizer(self):
        self.optimizer = get_optimizer(self.cfg.hyperparams, self.model)
        self.scheduler = get_scheduler(self.cfg.hyperparams, self.optimizer)

    def train(
        self,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_val: np.ndarray = None,
        y_val: np.ndarray = None,
        time_limit: float = None,
    ):
        time_start = time.time()
        if self.optimizer is None:
            self.reset_optimizer()
        # FIXME: Figure out best way to seed model
        rng = self.set_random_seed()
        use_val = x_val is not None

        checkpoint = Checkpoint(save_best=self.use_best_epoch, in_memory=True)

        self.preprocessor.fit(x_train, y_train)
        x_train = self.preprocessor.transform(x_train)

        if use_val:
            x_val = self.preprocessor.transform(x_val)
        self.y_transformer = create_y_transformer(y_train, self.cfg.task)

        dataset_train_generator = DatasetFinetuneGenerator(
            self.cfg,
            x=x_train,
            y=self.y_transformer.transform(y_train),
            task=self.cfg.task,
            max_samples_support=self.cfg.hyperparams["max_samples_support"],
            max_samples_query=self.cfg.hyperparams["max_samples_query"],
            split=0.8,
            random_state=rng,
        )

        if use_val:
            dataset_valid = DatasetFinetune(
                self.cfg,
                x_support=x_train,
                y_support=self.y_transformer.transform(y_train),
                x_query=x_val,
                y_query=y_val,
                max_samples_support=self.cfg.hyperparams["max_samples_support"],
                max_samples_query=self.cfg.hyperparams["max_samples_query"],
            )
            loader_valid = self.make_loader(dataset_valid, training=False)
        else:
            loader_valid = None

        if use_val and self.use_best_epoch:
            checkpoint.reset()

        max_epochs = self.cfg.hyperparams["max_epochs"]

        epoch = 0
        if max_epochs != 0 and use_val:
            metrics_valid = self.test_epoch(loader_valid, y_val)

            log_msg = f"Epoch 000"
            if self.compute_train_metrics:
                log_msg += f" | Train error: -.---- | Train score: -.---- |"
            if metrics_valid is not None:
                log_msg += f" | Val error: {metrics_valid.loss:.4f} | Val score: {metrics_valid.score:.4f}"

            logger.log(15, log_msg)
            if self.use_best_epoch:
                checkpoint(self.model, metrics_valid.loss, epoch=0)

            if time_limit is not None:
                time_cur = time.time()
                time_elapsed = time_cur - time_start
                time_left = time_limit - time_elapsed
                if time_left < (time_elapsed * 3 + 3):
                    # Fine-tuning an epoch will take longer than this, so triple the time required
                    logger.log(15, "Early stopping due to running out of time...")
                    max_epochs = 0

        for epoch in range(1, max_epochs + 1):
            dataset_train = next(dataset_train_generator)
            loader_train = self.make_loader(dataset_train, training=True)

            metrics_train = self.train_epoch(loader_train, return_metrics=self.compute_train_metrics)
            if use_val:
                metrics_valid = self.test_epoch(loader_valid, y_val)
            else:
                metrics_valid = None

            log_msg = f"Epoch {epoch:03d}"
            if metrics_train is not None:
                log_msg += f" | Train error: {metrics_train.loss:.4f} | Train score: {metrics_train.score:.4f}"
            if metrics_valid is not None:
                log_msg += f" | Val error: {metrics_valid.loss:.4f} | Val score: {metrics_valid.score:.4f}"

            logger.log(15, log_msg)
            if metrics_valid is not None:
                if self.use_best_epoch:
                    checkpoint(self.model, metrics_valid.loss, epoch=epoch)

                self.early_stopping(metrics_valid.loss)
                if self.early_stopping.we_should_stop():
                    logger.info("Early stopping")
                    break
                self.scheduler.step(
                    metrics_valid.loss
                )  # TODO: Make scheduler work properly during refit with no val data, to mimic scheduler in OG fit

            if time_limit is not None:
                time_cur = time.time()
                time_elapsed = time_cur - time_start

                time_per_epoch = time_elapsed / epoch
                time_left = time_limit - time_elapsed
                if time_left < (time_per_epoch + 3):
                    logger.log(15, "Early stopping due to running out of time...")
                    break

        if use_val and self.use_best_epoch and checkpoint.best_model is not None:
            # TODO: Can do a trick: Skip saving and loading best epoch if best epoch is the final epoch, will save around ~0.5 seconds
            self.best_epoch = checkpoint.best_epoch
            self.model.load_state_dict(checkpoint.load())
        else:
            self.best_epoch = epoch

    def minimize_for_inference(self):
        # delete unnecessary objects for inference
        self.optimizer = None
        self.scheduler = None

    def train_epoch(
        self, dataloader: torch.utils.data.DataLoader, return_metrics: bool = False
    ) -> PredictionMetrics | None:
        """

        Parameters
        ----------
        dataloader
        return_metrics: bool = False
            If True, will calculate and return metrics on the train data.
            Note that this can slow down training speed by >10%.

        Returns
        -------

        """
        assert self.optimizer is not None
        self.model.train()

        if return_metrics:
            output_tracker = TrackOutput()
        else:
            output_tracker = None

        for batch in dataloader:
            self.optimizer.zero_grad()

            x_support = batch["x_support"].to(self.cfg.device)
            y_support = batch["y_support"].to(self.cfg.device)
            x_query = batch["x_query"].to(self.cfg.device)
            y_query = batch["y_query"].to(self.cfg.device)

            if self.cfg.task == Task.REGRESSION:
                x_support, y_support, x_query, y_query = (
                    x_support.float(),
                    y_support.float(),
                    x_query.float(),
                    y_query.float(),
                )

            y_hat = self.model(x_support, y_support, x_query)

            if self.cfg.task == Task.REGRESSION:
                y_hat = y_hat[0, :, 0]
            else:
                y_hat = y_hat[0, :, : self.n_classes]

            y_query = y_query[0, :]

            loss = self.loss(y_hat, y_query)
            loss.backward()
            self.optimizer.step()

            if return_metrics:
                output_tracker.update(einops.asnumpy(y_query), einops.asnumpy(y_hat))

        if return_metrics:
            y_true, y_pred = output_tracker.get()
            y_pred = self.y_transformer.inverse_transform(y_pred)
            prediction_metrics = PredictionMetrics.from_prediction(
                y_pred, y_true, self.cfg.task, metric=self.stopping_metric
            )
            return prediction_metrics
        else:
            return None

    def test_epoch(self, dataloader: torch.utils.data.DataLoader, y_test: np.ndarray) -> PredictionMetrics:
        # FIXME: test_epoch might be better if it uses the for loop logic with n_ensembles
        y_hat = self.predict_epoch(dataloader)
        y_hat_finish = self.y_transformer.inverse_transform(y_hat)

        prediction_metrics = PredictionMetrics.from_prediction(
            y_hat_finish, y_test, self.cfg.task, metric=self.stopping_metric
        )
        return prediction_metrics

    def _get_memory_size(self) -> int:
        import gc
        import pickle
        import sys

        gc.collect()  # Try to avoid OOM error
        return sys.getsizeof(pickle.dumps(self, protocol=4))

    def predict(self, x_support: np.ndarray, y_support: np.ndarray, x_query: np.ndarray) -> np.ndarray:
        """
        Give a prediction for the query set.
        """

        x_support = self.preprocessor.transform(x_support)
        x_query = self.preprocessor.transform(x_query)

        dataset = DatasetFinetune(
            self.cfg,
            x_support=x_support,
            y_support=self.y_transformer.transform(y_support),
            x_query=x_query,
            y_query=None,
            max_samples_support=self.cfg.hyperparams["max_samples_support"],
            max_samples_query=self.cfg.hyperparams["max_samples_query"],
        )

        loader = self.make_loader(dataset, training=False)

        y_hat_ensembles = []

        for _ in range(self.cfg.hyperparams["n_ensembles"]):
            y_hat = self.predict_epoch(loader)
            y_hat_ensembles.append(y_hat)

        y_hat_ensembled = sum(y_hat_ensembles) / self.cfg.hyperparams["n_ensembles"]
        y_hat_finish = self.y_transformer.inverse_transform(y_hat_ensembled)

        return y_hat_finish

    def predict_epoch(self, dataloader: torch.utils.data.DataLoader) -> np.ndarray:
        """
        Returns the predictions for the data in the dataloader.
        The predictions are in the original state as they come from the model, i.e. not transformed.
        """

        self.model.eval()

        y_hat_list = []

        with torch.no_grad():
            for batch in dataloader:
                x_support = batch["x_support"].to(self.cfg.device)
                y_support = batch["y_support"].to(self.cfg.device)
                x_query = batch["x_query"].to(self.cfg.device)

                if self.cfg.task == Task.REGRESSION:
                    y_support = y_support.float()

                y_hat = self.model(x_support, y_support, x_query)

                if self.cfg.task == Task.REGRESSION:
                    y_hat = y_hat[0, :, 0]
                else:
                    y_hat = y_hat[0, :, : self.n_classes]

                y_hat_list.append(einops.asnumpy(y_hat))

        y_hat = np.concatenate(y_hat_list, axis=0)
        return y_hat

    def make_loader(self, dataset, training):
        return torch.utils.data.DataLoader(
            dataset,
            batch_size=1,
            shuffle=training,
            pin_memory=True,
            num_workers=0,
            drop_last=False,
            collate_fn=CollatorWithPadding(pad_to_n_support_samples=None),
        )



from abc import ABC, abstractmethod

import torch
import torch.nn as nn


class BaseModel(nn.Module, ABC):
    def __init__(self):
        super().__init__()

    def init_weights(self):
        """Initialize model weights."""
        pass

    @abstractmethod
    def forward(self, x_support: torch.Tensor, y_support: torch.Tensor, x_query: torch.Tensor, **kwargs):
        """Forward pass for the model."""
        pass



from __future__ import annotations

import logging

from autogluon.core.models import AbstractModel

logger = logging.getLogger(__name__)


# TODO: Add type hints once torch is a required dependency
class AbstractTorchModel(AbstractModel):
    """
    .. versionadded:: 1.5.0
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.device = None
        self.device_train = None

    def suggest_device_infer(self, verbose: bool = False) -> str:
        import torch

        # Put the model on the same device it was trained on (GPU/MPS) if it is available; otherwise use CPU
        if self.device_train is None:
            original_device_type = None  # skip update because no device is recorded
        elif isinstance(self.device_train, str):
            original_device_type = self.device_train
        else:
            original_device_type = self.device_train.type
        if original_device_type is None:
            # fallback to CPU
            device = torch.device("cpu")
        elif "cuda" in original_device_type:
            # cuda: nvidia GPU
            device = torch.device(original_device_type if torch.cuda.is_available() else "cpu")
        elif "mps" in original_device_type:
            # mps: Apple Silicon
            device = torch.device(original_device_type if torch.backends.mps.is_available() else "cpu")
        else:
            device = torch.device(original_device_type)

        if verbose and (original_device_type != device.type):
            logger.log(
                15,
                f"Model is trained on {original_device_type}, but the device is not available - "
                f"loading on {device.type}...",
            )

        return device.type

    @classmethod
    def to_torch_device(cls, device: str):
        import torch

        return torch.device(device)

    def get_device(self) -> str:
        """
        Returns torch.device(...) of the fitted model

        Requires implementation by the inheriting model class.
        Refer to overriding methods in existing models for reference implementations.
        """
        raise NotImplementedError

    def set_device(self, device: str):
        if not isinstance(device, str):
            device = device.type
        self.device = device
        self._set_device(device=device)

    def _set_device(self, device: str):
        """
        Sets the device for the inner model object.

        Requires implementation by the inheriting model class.
        Refer to overriding methods in existing models for reference implementations.

        If your model does not need to edit inner model object details, you can simply make the logic `pass`.
        """
        raise NotImplementedError

    def _post_fit(self, **kwargs):
        super()._post_fit(**kwargs)
        if self._get_class_tags().get("can_set_device", False):
            self.device_train = self.get_device()
            self.device = self.device_train
        return self

    def save(self, path: str = None, verbose=True) -> str:
        """
        Need to set device to CPU to be able to load on a non-GPU environment
        """
        reset_device = False
        og_device = self.device

        # Save on CPU to ensure the model can be loaded without GPU
        if self.is_fit():
            device_save = self._get_class_tags().get("set_device_on_save_to", None)
            if device_save is not None:
                self.set_device(device=device_save)
                reset_device = True
        path = super().save(path=path, verbose=verbose)
        # Put the model back to the device after the save
        if reset_device:
            self.set_device(device=og_device)
        return path

    @classmethod
    def load(cls, path: str, reset_paths=True, verbose=True):
        """
        Loads the model from disk to memory.
        The loaded model will be on the same device it was trained on (cuda/mps);
        if the device is not available (trained on GPU, deployed on CPU), then `cpu` will be used.

        Parameters
        ----------
        path : str
            Path to the saved model, minus the file name.
            This should generally be a directory path ending with a '/' character (or appropriate path separator value depending on OS).
            The model file is typically located in os.path.join(path, cls.model_file_name).
        reset_paths : bool, default True
            Whether to reset the self.path value of the loaded model to be equal to path.
            It is highly recommended to keep this value as True unless accessing the original self.path value is important.
            If False, the actual valid path and self.path may differ, leading to strange behaviour and potential exceptions if the model needs to load any other files at a later time.
        verbose : bool, default True
            Whether to log the location of the loaded file.

        Returns
        -------
        model : cls
            Loaded model object.
        """
        model = super().load(path=path, reset_paths=reset_paths, verbose=verbose)

        # Put the model on the same device it was trained on (GPU/MPS) if it is available; otherwise use CPU
        if model.is_fit() and model._get_class_tags().get("set_device_on_load", False):
            device = model.suggest_device_infer(verbose=verbose)
            model.set_device(device=device)

        return model

    @classmethod
    def _class_tags(cls):
        return {
            "can_set_device": True,
            "set_device_on_save_to": "cpu",
            "set_device_on_load": True,
        }



from autogluon.common.space import Categorical, Real


def get_default_searchspace(problem_type, num_classes=None):
    spaces = {
        "C": Real(lower=0.1, upper=1e3, default=1),
        "proc.skew_threshold": Categorical(0.99, None),
        "penalty": Categorical("L2", "L1"),
    }
    return spaces
